import { ModuleControl } from "queenjs";
import { reactive } from "vue";
import { EditorModule } from "../../module";
import { DesignComp } from "../../objects/DesignTemp/DesignComp";
import * as transforms from "./transforms";

type TransHandle = (this: TransferCtrl, e: MouseEvent) => void;

export type TransCreateFn = (...args: any[]) => {
  mousedown?: TransHandle;
  mousemove: TransHandle;
  mouseup: TransHandle;
};

export class TransferCtrl extends ModuleControl<EditorModule> {
  compEl!: HTMLElement;
  pageEl!: HTMLElement;
  currComp!: DesignComp;
  currTransfer!: ReturnType<TransCreateFn>;
  currObserver?: MutationObserver;
  transforms = transforms;

  transEvent = {
    startX: 0,
    startY: 0,
    offsetX: 0,
    offsetY: 0,
    width: 0,
    height: 0,
  };
  transferStyle = reactive({
    top: "",
    left: "",
    width: "",
    height: "",
    transform: "",
  });

  init(pageEl: HTMLElement) {
    this.currComp = this.module.store.currComp;
    this.compEl = this.currComp.$el;
    this.pageEl = pageEl;
    this.observe();
  }

  mousedown(e: MouseEvent, type: keyof typeof transforms) {
    const { currComp } = this.store;
    this.transEvent = {
      startX: e.clientX,
      startY: e.clientY,
      offsetX: 0,
      offsetY: 0,
      width: currComp.layout.size?.[0] || this.compEl.clientWidth * 2,
      height: currComp.layout.size?.[1] || this.compEl.clientHeight * 2,
    };
    this.currTransfer = this.transforms[type]();
    document.addEventListener("mousemove", this.mousemove);
    document.addEventListener("mouseup", this.mouseup);
    this.currTransfer.mousedown?.call(this, e);
  }
  private mousemove = (e: MouseEvent) => {
    const { transEvent } = this;
    transEvent.offsetX = e.clientX - transEvent.startX;
    transEvent.offsetY = e.clientY - transEvent.startY;
    this.currTransfer.mousemove.call(this, e);
  };
  private mouseup = (e: MouseEvent) => {
    document.removeEventListener("mousemove", this.mousemove);
    document.removeEventListener("mouseup", this.mouseup);
    this.currTransfer.mouseup.call(this, e);
  };

  observe() {
    if (this.currObserver) {
      this.currObserver.disconnect();
    }
    this.initStyle();
    this.currObserver = new MutationObserver((mutations) => {
      mutations.forEach((mutation) => {
        if (
          mutation.type === "childList" ||
          (mutation.type === "attributes" && mutation.attributeName === "style")
        ) {
          this.initStyle();
        }
      });
    });
    this.currObserver.observe(this.compEl, {
      attributes: true,
      childList: true,
      subtree: true,
      characterData: true,
    });
  }

  initStyle() {
    const rect = this.compEl.getBoundingClientRect();
    const pageRect = this.pageEl.getBoundingClientRect();
    this.transferStyle.width = rect.width + "px";
    this.transferStyle.height = rect.height + "px";
    this.transferStyle.top = rect.top - pageRect.top + "px";
    this.transferStyle.left = rect.left - pageRect.left + "px";
  }

  resetStyle() {
    Object.keys(this.transferStyle).forEach((key) => {
      (this.transferStyle as any)[key] = "";
    });
  }

  destroy() {
    this.currObserver?.disconnect();
    this.resetStyle();
  }
}
